module DE2_TOP (
		// Clock Input
		input 	      CLOCK_27, // 27 MHz
		input 	      CLOCK_50, // 50 MHz
		input 	      EXT_CLOCK, // External Clock
		// Push Button
		input [3:0]   KEY, // Pushbutton[3:0]
		// DPDT Switch
		input [17:0]  SW, // Toggle Switch[17:0]
		// 7-SEG Display
		output [6:0]  HEX0, // Seven Segment Digit 0
		output [6:0]  HEX1, // Seven Segment Digit 1
		output [6:0]  HEX2, // Seven Segment Digit 2
		output [6:0]  HEX3, // Seven Segment Digit 3
		output [6:0]  HEX4, // Seven Segment Digit 4
		output [6:0]  HEX5, // Seven Segment Digit 5
		output [6:0]  HEX6, // Seven Segment Digit 6
		output [6:0]  HEX7, // Seven Segment Digit 7
		// LED
		output [8:0]  LEDG, // LED Green[8:0]
		output [17:0] LEDR, // LED Red[17:0]
		// UART
		output 	      UART_TXD, // UART Transmitter
		input 	      UART_RXD, // UART Receiver
		// IRDA
		output 	      IRDA_TXD, // IRDA Transmitter
		input 	      IRDA_RXD, // IRDA Receiver
		// SDRAM Interface
		inout [15:0]  DRAM_DQ, // SDRAM Data bus 16 Bits
		output [11:0] DRAM_ADDR, // SDRAM Address bus 12 Bits
		output 	      DRAM_LDQM, // SDRAM Low-byte Data Mask 
		output 	      DRAM_UDQM, // SDRAM High-byte Data Mask
		output 	      DRAM_WE_N, // SDRAM Write Enable
		output 	      DRAM_CAS_N, // SDRAM Column Address Strobe
		output 	      DRAM_RAS_N, // SDRAM Row Address Strobe
		output 	      DRAM_CS_N, // SDRAM Chip Select
		output 	      DRAM_BA_0, // SDRAM Bank Address 0
		output 	      DRAM_BA_1, // SDRAM Bank Address 0
		output 	      DRAM_CLK, // SDRAM Clock
		output 	      DRAM_CKE, // SDRAM Clock Enable
		// Flash Interface
		inout [7:0]   FL_DQ, // FLASH Data bus 8 Bits
		output [21:0] FL_ADDR, // FLASH Address bus 22 Bits
		output 	      FL_WE_N, // FLASH Write Enable
		output 	      FL_RST_N, // FLASH Reset
		output 	      FL_OE_N, // FLASH Output Enable
		output 	      FL_CE_N, // FLASH Chip Enable
		// SRAM Interface
		inout [15:0]  SRAM_DQ, // SRAM Data bus 16 Bits
		output [17:0] SRAM_ADDR, // SRAM Address bus 18 Bits
		output 	      SRAM_UB_N, // SRAM High-byte Data Mask 
		output 	      SRAM_LB_N, // SRAM Low-byte Data Mask 
		output 	      SRAM_WE_N, // SRAM Write Enable
		output 	      SRAM_CE_N, // SRAM Chip Enable
		output 	      SRAM_OE_N, // SRAM Output Enable
		// ISP1362 Interface
		inout [15:0]  OTG_DATA, // ISP1362 Data bus 16 Bits
		output [1:0]  OTG_ADDR, // ISP1362 Address 2 Bits
		output 	      OTG_CS_N, // ISP1362 Chip Select
		output 	      OTG_RD_N, // ISP1362 Write
		output 	      OTG_WR_N, // ISP1362 Read
		output 	      OTG_RST_N, // ISP1362 Reset
		output 	      OTG_FSPEED, // USB Full Speed, 0 = Enable, Z = Disable
		output 	      OTG_LSPEED, // USB Low Speed,  0 = Enable, Z = Disable
		input 	      OTG_INT0, // ISP1362 Interrupt 0
		input 	      OTG_INT1, // ISP1362 Interrupt 1
		input 	      OTG_DREQ0, // ISP1362 DMA Request 0
		input 	      OTG_DREQ1, // ISP1362 DMA Request 1
		output 	      OTG_DACK0_N, // ISP1362 DMA Acknowledge 0
		output 	      OTG_DACK1_N, // ISP1362 DMA Acknowledge 1
		// LCD Module 16X2
		inout [7:0]   LCD_DATA, // LCD Data bus 8 bits
		output 	      LCD_ON, // LCD Power ON/OFF
		output 	      LCD_BLON, // LCD Back Light ON/OFF
		output 	      LCD_RW, // LCD Read/Write Select, 0 = Write, 1 = Read
		output 	      LCD_EN, // LCD Enable
		output 	      LCD_RS, // LCD Command/Data Select, 0 = Command, 1 = Data
		// SD Card Interface
		inout 	      SD_DAT, // SD Card Data
		inout 	      SD_DAT3, // SD Card Data 3
		inout 	      SD_CMD, // SD Card Command Signal
		output 	      SD_CLK, // SD Card Clock
		// I2C
		inout 	      I2C_SDAT, // I2C Data
		output 	      I2C_SCLK, // I2C Clock
		// PS2
		input 	      PS2_DAT, // PS2 Data
		input 	      PS2_CLK, // PS2 Clock
		// USB JTAG link
		input 	      TDI, // CPLD -> FPGA (data in)
		input 	      TCK, // CPLD -> FPGA (clk)
		input 	      TCS, // CPLD -> FPGA (CS)
		output 	      TDO, // FPGA -> CPLD (data out)
		// VGA
		output 	      VGA_CLK, // VGA Clock
		output 	      VGA_HS, // VGA H_SYNC
		output 	      VGA_VS, // VGA V_SYNC
		output 	      VGA_BLANK, // VGA BLANK
		output 	      VGA_SYNC, // VGA SYNC
		output [9:0]  VGA_R, // VGA Red[9:0]
		output [9:0]  VGA_G, // VGA Green[9:0]
		output [9:0]  VGA_B, // VGA Blue[9:0]
		// Ethernet Interface
		inout [15:0]  ENET_DATA, // DM9000A DATA bus 16Bits
		output 	      ENET_CMD, // DM9000A Command/Data Select, 0 = Command, 1 = Data
		output 	      ENET_CS_N, // DM9000A Chip Select
		output 	      ENET_WR_N, // DM9000A Write
		output 	      ENET_RD_N, // DM9000A Read
		output 	      ENET_RST_N, // DM9000A Reset
		input 	      ENET_INT, // DM9000A Interrupt
		output 	      ENET_CLK, // DM9000A Clock 25 MHz
		// Audio CODEC
		inout 	      AUD_ADCLRCK, // Audio CODEC ADC LR Clock
		input 	      AUD_ADCDAT, // Audio CODEC ADC Data
		inout 	      AUD_DACLRCK, // Audio CODEC DAC LR Clock
		output 	      AUD_DACDAT, // Audio CODEC DAC Data
		inout 	      AUD_BCLK, // Audio CODEC Bit-Stream Clock
		output 	      AUD_XCK, // Audio CODEC Chip Clock
		// TV Decoder
		input [7:0]   TD_DATA, // TV Decoder Data bus 8 bits
		input 	      TD_HS, // TV Decoder H_SYNC
		input 	      TD_VS, // TV Decoder V_SYNC
		output 	      TD_RESET, // TV Decoder Reset
		// GPIO
		inout [35:0]  GPIO_0, // GPIO Connection 0
		inout [35:0]  GPIO_1       // GPIO Connection 1
		);

   //Turn off all displays.
   assign HEX0 = 7'h7F;
   assign HEX1 = 7'h7F;
   assign HEX2 = 7'h7F;
   assign HEX3 = 7'h7F;
   assign HEX4 = 7'h7F;
   assign HEX5 = 7'h7F;
   assign HEX6 = 7'h7F;
   assign HEX7 = 7'h7F;
   //   assign LEDR = 18'h0;
   // assign LEDG = 9'h0;
   
   //Set all GPIO to tri-state.
   assign GPIO_0 = 36'hzzzzzzzzz;
   assign GPIO_1 = 36'hzzzzzzzzz;

   //Disable audio codec.
   //assign AUD_DACDAT = 1'b0;
   //assign AUD_XCK    = 1'b0;

   //Disable DRAM.
   assign DRAM_ADDR  = 12'h0;
   assign DRAM_BA_0  = 1'b0;
   assign DRAM_BA_1  = 1'b0;
   assign DRAM_CAS_N = 1'b1;
   assign DRAM_CKE   = 1'b0;
   assign DRAM_CLK   = 1'b0;
   assign DRAM_CS_N  = 1'b1;
   assign DRAM_DQ    = 16'hzzzz;
   assign DRAM_LDQM  = 1'b0;
   assign DRAM_RAS_N = 1'b1;
   assign DRAM_UDQM  = 1'b0;
   assign DRAM_WE_N  = 1'b1;

   //Disable Ethernet.
   assign ENET_CLK   = 1'b0;
   assign ENET_CS_N  = 1'b1;
   assign ENET_CMD   = 1'b0;
   assign ENET_DATA  = 16'hzzzz;
   assign ENET_RD_N  = 1'b1;
   assign ENET_RST_N = 1'b1;
   assign ENET_WR_N  = 1'b1;

   //Disable flash.
   assign FL_ADDR  = 22'h0;
   assign FL_CE_N  = 1'b1;
   assign FL_DQ    = 8'hzz;
   assign FL_OE_N  = 1'b1;
   assign FL_RST_N = 1'b1;
   assign FL_WE_N  = 1'b1;

   //Disable LCD.
   assign LCD_BLON = 1'b0;
   assign LCD_DATA = 8'hzz;
   assign LCD_EN   = 1'b0;
   assign LCD_ON   = 1'b0;
   assign LCD_RS   = 1'b0;
   assign LCD_RW   = 1'b0;

   //Disable OTG.
   assign OTG_ADDR    = 2'h0;
   assign OTG_CS_N    = 1'b1;
   assign OTG_DACK0_N = 1'b1;
   assign OTG_DACK1_N = 1'b1;
   assign OTG_FSPEED  = 1'b1;
   assign OTG_DATA    = 16'hzzzz;
   assign OTG_LSPEED  = 1'b1;
   assign OTG_RD_N    = 1'b1;
   assign OTG_RST_N   = 1'b1;
   assign OTG_WR_N    = 1'b1;

   //Disable SDRAM.
   assign SD_DAT = 1'bz;
   assign SD_CLK = 1'b0;

   //Disable SRAM.
   assign SRAM_ADDR = 18'h0;
   assign SRAM_CE_N = 1'b1;
   assign SRAM_DQ   = 16'hzzzz;
   assign SRAM_LB_N = 1'b1;
   assign SRAM_OE_N = 1'b1;
   assign SRAM_UB_N = 1'b1;
   assign SRAM_WE_N = 1'b1;

   //Disable VGA.
   /*
    assign VGA_CLK   = 1'b0;
    assign VGA_BLANK = 1'b0;
    assign VGA_SYNC  = 1'b0;
    assign VGA_HS    = 1'b0;
    assign VGA_VS    = 1'b0;
    assign VGA_R     = 10'h0;
    assign VGA_G     = 10'h0;
    assign VGA_B     = 10'h0;
    */

   //Disable all other peripherals.
   //assign I2C_SCLK = 1'b0;
   assign IRDA_TXD = 1'b0;
   //assign TD_RESET = 1'b0;
   assign TDO = 1'b0;
   assign UART_TXD = 1'b0;
   

   wire 		      VGA_CTRL_CLK;
   wire 		      AUD_CTRL_CLK;
   wire 		      DLY_RST;
   
   assign  TD_RESET  =  1'b1;  //  Allow 27 MHz
   assign  AUD_ADCLRCK  =  AUD_DACLRCK;
   assign  AUD_XCK    =  AUD_CTRL_CLK;

   Reset_Delay      r0  (  .iCLK(CLOCK_50),.oRESET(DLY_RST)  );
   
   VGA_Audio_PLL     p1  (  .areset(~DLY_RST),.inclk0(CLOCK_27),.c0(VGA_CTRL_CLK),.c1(AUD_CTRL_CLK),.c2(VGA_CLK)  );
   
   
   VGA_Controller    u1  (  //  Host Side
			    .iCursor_RGB_EN(4'b1111),
			    .oAddress(mVGA_ADDR),
			    .oCoord_X(Coord_X),
			    .oCoord_Y(Coord_Y),
			    .iCursor_X(SW[17:11]),
			    .iCursor_Y(SW[10:4]),
			    .iCursor_R(10'h3FF),
			    .iCursor_G(10'h3FF),
			    .iCursor_B(10'h0),
			    .iRed(mVGA_R),
			    .iGreen(mVGA_G),
			    .iBlue(mVGA_B),
			    //  VGA Side
			    .oVGA_R(VGA_R),
			    .oVGA_G(VGA_G),
			    .oVGA_B(VGA_B),
			    .oVGA_H_SYNC(VGA_HS),
			    .oVGA_V_SYNC(VGA_VS),
			    .oVGA_SYNC(VGA_SYNC),
			    .oVGA_BLANK(VGA_BLANK),
			    //  Control Signal
			    .iCLK(VGA_CTRL_CLK),
			    .iRST_N(DLY_RST)  );

   wire [9:0] 		      mVGA_R;        //memory output to VGA
   wire [9:0] 		      mVGA_G;
   wire [9:0] 		      mVGA_B;
   wire [18:0] 		      mVGA_ADDR;      //video memory address
   wire [9:0] 		      Coord_X, Coord_Y;  //display coods

   ////////////////////////////////////
   //DLA state machine variables
   wire 		      reset;
   reg [30:0] 		      x_rand;  //shift registers for random numbe

   wire 		      x_low_bit;
   assign x_low_bit = x_rand[27] ^ x_rand[30];
   
   reg 			      blue_en1;
   reg 			      green_en1;
   reg 			      red_en1;
   reg 			      blue_en2;
   reg 			      green_en2;
   reg 			      red_en2;
   
   //buffer for edges of VGA out
   //parameter COL_START = 10'd0;
   //parameter COL_END = 10'd639;
   //parameter ROW_START = 9'd0;
   //parameter ROW_END = 10'd479;
   
   ////////////////////////////////////
   /*From megaWizard:
    module vga_buffer (
    address_a, // use a for state machine
    address_b, // use b for VGA refresh
    clock_a,
    clock_b,
    data_a,
    data_b,
    wren_a,
    wren_b,
    q_a,
    q_b);*/
   // Show m4k on the VGA
   // -- use m4k a for state machine
   // -- use m4k b for VGA refresh
   //wire 		      mem_bit ; //current data from m4k to VGA
   //wire 			      disp_bit ; // registered data from m4k to VGA
  // wire 		      state_bit ; // current data from m4k to state machine
   //reg 			      we ; // write enable for a
   //reg [18:0] 		      addr_reg ; // for a
   //reg 			      data_reg ; // for a
   /*
   vga_buffer display(
		      .address_a (addr_reg) , 
		      .address_b (display_address), // vga current address
		      .clock_a (VGA_CTRL_CLK),
		      .clock_b (VGA_CTRL_CLK),
		      .data_a (data_reg),
		      .data_b (1'b0), // never write on port b
		      .wren_a (we),
		      .wren_b (1'b0), // never write on port b
		      .q_a (state_bit),
		      .q_b (mem_bit) ); // data used to update VGA

*/
       // make the color
   assign  mVGA_R = {10{disp_bit}}*red_en2 - {10{disp_bit}}*Coord_Y*10'd2*red_en1;
   assign  mVGA_G = {10{disp_bit}}*green_en2 - {10{disp_bit}}*Coord_Y*10'd2*green_en1;
   assign  mVGA_B = {10{disp_bit}}*blue_en2 - {10{disp_bit}}*Coord_Y*10'd2*blue_en1;

   // DLA state machine
   assign reset = ~KEY[0];

   wire 		      run;
   assign run = 1'd1;
   
   GridController controller(.iVGA_request_x(Coord_X),
			     .iVGA_request_y(Coord_Y),
			     .irun(run),
			     .ioverride_x(override_x_from_cpu),
			     .ioverride_y(override_y_from_cpu),
			     .ioverride_data(override_data_from_cpu),
			     .ioverride_clk(override_clk_from_cpu),
			     .oVGA_value(disp_bit),
			     .clk(VGA_CTRL_CLK),
			     .mem_clk(CLOCK_50),
			     .reset(reset),
			     .random_reset(SW[2:0]),
			     .iframes_to_skip(SW[17:10]));


   //Colors:
    always @ (posedge VGA_CTRL_CLK) //VGA_CTRL_CLK
     begin
	// register the m4k output for better timing on VGA
	//disp_bit <= mem_bit;
	x_rand <= {x_rand[29:0] , x_low_bit} ;
	
	if (reset) //synch reset assumes KEY0 is held down 1/60 second
	  begin
	     if (x_rand == 31'h0)
               x_rand <= 31'h55555555;
	     // Pick Random Colors:
	     red_en1 <= x_rand[0];
	     red_en2 <= x_rand[1];
	     green_en1 <= x_rand[2];
	     green_en2 <= x_rand[3];
	     blue_en1 <= x_rand[4];
	     blue_en2 <= x_rand[5];
	  end // if (reset)
     end // always @ (posedge VGA_CTRL_CLK)


   //////////////////////////////////////////
   // CPU
   //////////////////////////////////////////

   wire [9:0] cursor_x_from_cpu, cursor_y_from_cpu;
   wire [9:0] override_x_from_cpu, override_y_from_cpu;
   wire       override_clk_from_cpu;
   wire       override_data_from_cpu;
   
   
   
   /*
   //state names
   parameter init = 4'd0, 
     update_load1 = 4'd1,
     update_load2 = 4'd2,
     update_load3 = 4'd3,
     update_load4 = 4'd4,
     update_load5 = 4'd5,
     update_load6 = 4'd6,
     update_calculate = 4'd7,
     wait_for_vga_start = 4'd8,
     wait_for_next_pixel = 4'd9,
     save_result = 4'd10,
     save_result2 = 4'd11;
   
   always @ (negedge VGA_CTRL_CLK)
     begin
	// register the m4k output for better timing on VGA
	// negedge seems to work better than posedge
	disp_bit <= mem_bit;
     end
    
   wire prev_row_out;
   reg [0:0] 		      prev_row_data;
   reg [9:0] prev_row_address;
   reg [0:0] prev_row_we;
   
   pixel_row prev_row(.out(prev_row_out),
		      .address(prev_row_address),
		      .data(prev_row_data),
		      .we(prev_row_we),
		      .clk(CLOCK_50));

   GridController controller(.iVGA_request_x(Coord_X),
			     .iVGA_request_y(Coord_Y),
			     .oVGA_value(disp_bit),
			     .clk(VGA_CTRL_CLK));
   
   
   
   always @ (posedge VGA_CTRL_CLK) //VGA_CTRL_CLK
     begin
	// register the m4k output for better timing on VGA
	//disp_bit <= mem_bit;
	x_rand <= {x_rand[29:0] , x_low_bit} ;
	
	if (reset) //synch reset assumes KEY0 is held down 1/60 second
	  begin
	     //clear the screen
	     addr_reg <= display_address;
	     we <= 1'b1;          //write some memory
	     data_reg <= x_rand[0];        //write random bits
	     //clear prev_row
	     prev_row_we <= 1'b1;
	     prev_row_data <= 1'b0;
	     prev_row_address <= display_address[9:0];
	     //init random number generators to alternating bits
	     if (x_rand == 31'h0)
               x_rand <= 31'h55555555;
	     y_rand <= 29'h55555555;
	     state <= init;  //first state in regular state machine 
	     // Pick Random Colors:
	     red_en1 <= x_rand[0];
	     red_en2 <= x_rand[1];
	     green_en1 <= x_rand[2];
	     green_en2 <= x_rand[3];
	     blue_en1 <= x_rand[4];
	     blue_en2 <= x_rand[5];
	  end // if (reset)
	/*
	else
	  begin
	     case(state)
               init:
		 begin
		    prev_row_we <= 1'b0;
		    state <= wait_for_vga_start;
		 end // case: init

	       wait_for_vga_start:
		 begin
		    we <= 1'b0;
		    if(Coord_X[9:0] == 10'd0 && Coord_Y[8:0] == 9'd0)
		      state <= update_load1;
		 end

	       update_load1:
		 begin
		    // House keeping:
		    prev_pixel <= display_address;
		    lower_left_set <= 1'b0;
		    lower_center_set <= 1'b0;
		    lower_right_set <= 1'b0;

		    // Retrive saved previous
		    if(Coord_X[9:0] == 10'd0)
		      middle_left <= 1'b0;
		    else
		      middle_left <= result;
		    
		    // Pipeline middle_right
		    we <= 1'd0;
		    if(Coord_X[9:0] == 10'd639)
		      begin
			 middle_right_set <= 1'b1;
			 middle_right <= 1'b0;
		      end
		    else
		      begin
			 middle_right_set <= 1'b0;
			 addr_reg <= display_address + 'd1;
		      end

		    //Pipeline upper_left
		    prev_row_address <= Coord_X[9:0] - 10'd1;
		    prev_row_we <= 1'b0;

		    state <= update_load2;
		 end // case: update_load1
	       	       
		update_load2:
		  begin
		     //Pipeline lower_left
		     if(Coord_X[9:0] == 10'd0 || Coord_Y[8:0] == 9'd479) begin
			lower_left_set <= 1'b1;
			lower_left <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd639;
		     
		     //Pipeline upper_center
		     prev_row_address <= Coord_X[9:0];

		     //Save upper_left from Pipeline
		     upper_left <= prev_row_out;

		     state <= update_load3;
		  end // case: update_load2
	       
		update_load3:
		  begin
		     //Pipeline lower_center
		     if(Coord_Y[8:0] == 9'd479) begin
			lower_center_set <= 1'b1;
			lower_center <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd640;

		     //Save middle_right from pipeline
		     if(!middle_right_set)
		       middle_right <= state_bit;
		     
		     //Pipeline upper_right
		     prev_row_address <= Coord_X[9:0] + 10'd1;

		     //Save upper_center from Pipeline
		     upper_center <= prev_row_out;

		     state <= update_load4;
		  end // case: update_load3
	       
		update_load4:
		  begin
		     //load center:
		     middle_center <= disp_bit;
		     
		     //Pipeline lower_right
		     if(Coord_Y[8:0] == 9'd479 || Coord_X[9:0] == 10'd639) begin
			lower_right_set <= 1'b1;
			lower_right <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd641;

		     //Save lower_left from pipeline
		     if(!lower_left_set)
		       lower_left <= state_bit;
		     
		     //Save upper_right from Pipeline
		     upper_right <= prev_row_out;

		     state <= update_load5;
		  end // case: update_load4
	       	       
		update_load5:
		  begin
		     //Save lower_center from pipeline
		     if(!lower_center_set)
		       lower_center <= state_bit;

		     state <= update_load6;
		  end // case: update_load5
	       	       
		update_load6:
		  begin
		     //Save lower_center from pipeline
		     if(!lower_right_set)
		       lower_right <= state_bit;

		     state <= update_calculate;
		  end // case: update_load6
	       
	       update_calculate:
		 begin
		    if(neighbor_count == 'd3)
		      result <= 1'd1;
		    else if(neighbor_count == 'd2 && middle_center)
		      result <= 1'd1;
		    else
		      result <= 1'd0;
		    state <= wait_for_next_pixel;

		    // Also, update prev_row:
		    prev_row_we <= 1'b1;
		    prev_row_data <= middle_left;
		    prev_row_address <= Coord_X[9:0] - 'd1;
		 end

	       wait_for_next_pixel:
		 begin
		    if(prev_pixel != display_address)
		      state <= save_result;
		 end

	       save_result:
		 begin
		    addr_reg <= prev_pixel;
		    we <= 1'b1;
		    data_reg <= result;
		    state <= save_result2;
		 end
	       
	       save_result2:
		 begin
		    if(KEY[3]) begin
		       if(Coord_X == 10'd0 && Coord_Y == 9'd0)
			 state <= wait_for_vga_start;
		       else
			 state <= update_load1;
		    end
		 end // case: save_result2
	     endcase // case (state)
	  end // else: !if(reset) 
     end // always @ (posedge VGA_CTRL_CLK) */
endmodule //top module

module GridController(iVGA_request_x,  
		      iVGA_request_y,  
		      irun,	       
		      ioverride_x,     
		      ioverride_y,     
		      ioverride_data,  
		      ioverride_clk,
		      mem_clk,
		      clk,	       
		      reset,	     
		      random_reset,
		      oVGA_value,      
		      orunning,
		      iframes_to_skip
		      );
   
   input [9:0] iVGA_request_x;  
   input [9:0] iVGA_request_y;  
   input       irun;	       
   input [9:0] ioverride_x;     
   input [9:0] ioverride_y;     
   input       ioverride_data;  
   input       ioverride_clk;   
   input       clk;	  
   input       mem_clk;
   input       reset;	       
   input [2:0] random_reset;      
   output      oVGA_value;      
   output reg  orunning;
   input [9:0] iframes_to_skip;

   parameter CELLBLOCKS = 80;

   // Interconnects between cell blocks
   wire left_to_right_interconnects[CELLBLOCKS+1][3] ;
   wire right_to_left_interconnects[CELLBLOCKS+1][3] ;
   wire VGA_outs[CELLBLOCKS];
   reg override_en;
   reg [9:0] row;
   reg [9:0] prev_row;
   reg 	     sync;
   reg [9:0] frames_skipped;
   reg 	     cellblock_reset;
   reg 	     override_data;
   reg [9:0] override_x;
   reg [9:0] override_y;
   
   genvar i;
   generate
      for (i=0; i < CELLBLOCKS; i=i+1) begin : GENLOOP
	    CellBlock block (.iVGA_request_x(iVGA_request_x[2:0]),
			     .iVGA_request_y(iVGA_request_y),
			     .irow(row),		 
			     .isync(sync),		 
			     .clk(clk),	
			     .mem_clk(mem_clk),
			     .ioverride_en(override_en && (override_x[9:3] == i)),
			     .ioverride_x(override_x[2:0]),
			     .ioverride_y(override_y),
			     .ioverride_data(override_data), 
			     .iupper_right((i == CELLBLOCKS-1) ? 1'b0 : left_to_right_interconnects[i+1][0]),   
			     .imiddle_right((i == CELLBLOCKS-1) ? 1'b0 : left_to_right_interconnects[i+1][1]),	 
			     .ilower_right((i == CELLBLOCKS-1) ? 1'b0 : left_to_right_interconnects[i+1][2]),	 
			     .iupper_left((i == 0) ? 1'b0 : right_to_left_interconnects[i][0]),	 
			     .imiddle_left((i == 0) ? 1'b0 : right_to_left_interconnects[i][1]),	 
			     .ilower_left((i == 0) ? 1'b0 : right_to_left_interconnects[i][2]),            	 
			     .oVGA_value(VGA_outs[i]),	 
			     .oupper_right(right_to_left_interconnects[i+1][0]),  
			     .omiddle_right(right_to_left_interconnects[i+1][1]), 
			     .olower_right(right_to_left_interconnects[i+1][2]),  
			     .oupper_left(left_to_right_interconnects[i][0]),	       
			     .omiddle_left(left_to_right_interconnects[i][1]),	       
			     .olower_left(left_to_right_interconnects[i][2]),
			     .reset(cellblock_reset)
		    );
      end // block: GENLOOP
   endgenerate

   assign oVGA_value = VGA_outs[iVGA_request_x[9:3]];
      

   reg [30:0] random_reg;
   wire       random_low_bit;
   assign random_low_bit = random_reg[27] ^ random_reg[30];
   
   reg [3:0] state;
   
   parameter not_running = 4'd0,
     apply_override = 4'd1,
     wait_for_frame_start = 4'd2,
     frame_start = 4'd3,
     frame_start_skip = 4'd4,
     frame_start_init = 4'd5,
     wait_for_row_complete = 4'd6,
     execute_row = 4'd7,
     execute_row2 = 4'd8;
   
   always @(posedge clk)
     begin
	if(random_reg == 31'h0)
	  random_reg <= 31'h65656565;
	else
	  random_reg <= {random_reg[29:0] , random_low_bit};
	if(reset)
	  begin
	     if(random_reset == 3'd1) begin  // Fully Random Init
		override_data <= random_low_bit;
		override_en <= 1'b1;
		override_x <= iVGA_request_x;
		override_y <= iVGA_request_y;
	     end
	     else if(random_reset == 3'd2) begin // Half Random Init
		if(iVGA_request_y > 'd240)
		  override_data <= random_low_bit;
		else
		  override_data <= 1'b0;
		override_en <= 1'b1;
		override_x <= iVGA_request_x;
		override_y <= iVGA_request_y;
	     end
	     else if(random_reset == 3'd3) begin // Set Pattern Init
		if((iVGA_request_y == 'd240 && (iVGA_request_x == 'd320 || iVGA_request_x == 'd319 || iVGA_request_x == 'd321)) || (iVGA_request_y == 'd239 && iVGA_request_x == 'd320) || (iVGA_request_y == 'd241 && iVGA_request_x == 'd319))
		  override_data <= 1'd1;
		else
		  override_data <= 1'd0;
		override_en <= 1'b1;
		override_x <= iVGA_request_x;
		override_y <= iVGA_request_y;
	     end
	     else if(random_reset == 3'd4) begin  // Center Random Init
		if(iVGA_request_y > 'd180 && iVGA_request_y < 'd300 && iVGA_request_x > 'd240 && iVGA_request_x < 'd400)
		  override_data <= random_low_bit;
		else
		  override_data <= 1'b0;
		override_en <= 1'b1;
		override_x <= iVGA_request_x;
		override_y <= iVGA_request_y;
	     end
	     else begin  // Clear Screen Init
		cellblock_reset <= 1'd1;
		override_en <= 1'b0;
	     end
	     state <= not_running;
	     frames_skipped <= iframes_to_skip;
	  end
	else
	  begin
	     override_data <= ioverride_data;
	     override_x <= ioverride_x;
	     override_y <= ioverride_y;
	     case(state)
	       not_running:
		 begin
		    cellblock_reset <= 1'd0;
		    orunning <= 1'b0;
		    if(irun) begin
		       override_en <= 1'b0;
		       state <= wait_for_frame_start;
		    end
		    else if(ioverride_clk) begin
		       override_en <= 1'b1;
		       //last_overriden <= override_x[9:3];
		       state <= apply_override;
		    end
		    else
		       override_en <= 1'b0;
		 end // case: not_running

	       apply_override:
		 begin
		    if(!ioverride_clk) begin
		       override_en <= 1'b0;
		       state <= not_running;
		    end
		 end

	       wait_for_frame_start:
		 begin
		    orunning <= 1'b1;
		    if(!irun)
		      state <= not_running;
		    else if(iVGA_request_y == 10'd0)
		      state <= frame_start;
		 end

	       frame_start:
		 begin
		    if(frames_skipped < iframes_to_skip) begin
		       frames_skipped <= frames_skipped + 10'd1;
		       state <= frame_start_skip;
		    end
		    else begin
		       frames_skipped <= 10'd0;
		       state <= frame_start_init;
		    end
		 end // case: frame_start

	       frame_start_skip:
		 begin
		    if(iVGA_request_y != 10'd0)
		      state <= wait_for_frame_start;
		 end

	       frame_start_init:
		 begin
		    prev_row <= 10'd0;
		    state <= wait_for_row_complete;
		 end

	       wait_for_row_complete:
		 begin
		    if(prev_row != iVGA_request_y)
		      state <= execute_row;
		 end

	       execute_row:
		 begin
		    row <= prev_row; // row is what the cellblocks use
		    prev_row <= iVGA_request_y;
		    sync <= 1'b1;
		    state <= execute_row2;
		 end

	       execute_row2:
		 begin
		    sync <= 1'b0;
		    if(prev_row == 10'd0)
		      state <= wait_for_frame_start;
		    else
		      state <= wait_for_row_complete;
		 end
	     endcase // case (state)
	  end // else: !if(reset)
     end // always @ (posedge clk)
   
   // States:
   // not_running - check irun and overrides, react accordingly
   // apply_override - state while applying an override value
		    
   // wait_for_frame_start - waits for VGA_y == 'd0; also monitors irun
   // frame_start - Checks # of frames skipped.
   // frame_start_skip - waits for VGA_y != 'd0, then goes back to wait_for_frame_start
   // frame_start_init - does any initialization for new frame
   // wait_for_row_complete - Waits for VGA_y to increment
   // execute_row - Runs Cellblocks on newly completed row (raise sync to high)
   // execute_row2 - lower sync to low;  if VGA_y is currently 'd0, return to wait_for_frame_start instead of wait_for_row_complete.


   
endmodule // GridController

/* CellBlock: One block (set of adjacent columns) in the grid.
 note: Override should only be enabled while state machine is in waiting_for_sync.
       If this is not followed, It could cause corruption in the current cell by messing with pipelining.
 */
module CellBlock(iVGA_request_x,
		 iVGA_request_y,
		 irow,		
		 isync,		
		 clk,	
		 mem_clk,
		 ioverride_en,	
		 ioverride_x,	
		 ioverride_y,	
		 ioverride_data,
		 iupper_right,  
		 imiddle_right,	
		 ilower_right,	
		 iupper_left,	
		 imiddle_left,	
		 ilower_left,	
		 oVGA_value,	
		 oupper_right,  
		 omiddle_right,	
		 olower_right,	
		 oupper_left,	
		 omiddle_left,	
		 olower_left,
		 reset);

   parameter WIDTH = 8;
   parameter WIDTH_VALUE = 3'd7; //bus width should be WIDTH_WIRES
   parameter WIDTH_WIRES = 3; //3 wires holds 8 values (WIDTH)
   parameter HEIGHT = 9'd480;
   
   input [WIDTH_WIRES-1:0] iVGA_request_x;
   input [8:0] iVGA_request_y;
   input [8:0] irow;		
   input isync;		
   input clk;
   input mem_clk;
   input ioverride_en;	
   input [WIDTH_WIRES-1:0] ioverride_x;	
   input [8:0] ioverride_y;	
   input ioverride_data;
   input iupper_right;  
   input imiddle_right;	
   input ilower_right;	
   input iupper_left;	
   input imiddle_left;	
   input ilower_left;
   input reset;
   output oVGA_value;	
   output reg oupper_right;  
   output reg omiddle_right;	
   output reg olower_right;	
   output reg oupper_left;	
   output reg omiddle_left;	
   output reg olower_left;

   reg [4:0]             state;
   reg [8:0] 		 row;
   reg [WIDTH-1:0] 	 prev_row, original_row;
   reg [WIDTH_WIRES-1:0] col;
   reg 			 result;
   
   
   reg 			 mem_we, mem_data_in;
   reg [11:0] 		 mem_address;
   wire 		 mem_data_out;

   //Grid Locations
   reg upper_left;
   reg upper_center;
   reg upper_right;
   reg middle_left;
   reg middle_center;
   reg middle_right;
   reg lower_left;
   reg lower_center;
   reg lower_right;  

   wire [3:0] neighbor_count;
   assign neighbor_count = upper_left + 
			   upper_center + 
			   upper_right +  
			   middle_left +  
			   middle_right + 
			   lower_left +   
			   lower_center + 
			   lower_right;

   ///////////////////////////////
   // Begin Buffer System
   //////////////////////////////
   
   parameter row_change = 3'd0,
     buffer_setup = 3'd1,
     buffer_loop1 = 3'd2,
     buffer_loop2 = 3'd3,
     buffer_loop2b = 3'd5,
     buffer_loop3 = 3'd4;
      
   reg [7:0]  vga_row_buffer_current;
   reg [7:0]  vga_row_buffer_next;

   assign oVGA_value = vga_row_buffer_current[iVGA_request_x];

   reg [2:0]  buffer_state;
   reg [8:0]  buffer_last_vga_row;
   reg [8:0]  buffer_row;
   reg [2:0]  buffer_col;
   wire       buffer_mem_out;
   reg [11:0] buffer_address;
   always @(posedge clk)
     begin
	case(buffer_state)
	  row_change:
	    begin
	       vga_row_buffer_current <= vga_row_buffer_next;
	       buffer_last_vga_row <= iVGA_request_y[8:0];
	       buffer_state <= buffer_setup;
	    end
	  buffer_setup:
	    begin
	       if(buffer_last_vga_row == 9'd479)
		 buffer_row <= 9'd0;
	       else
		 buffer_row <= buffer_last_vga_row + 9'd1;
	       buffer_col <= 3'd0;
	       buffer_state <= buffer_loop1;
	    end
	  buffer_loop1:
	    begin
	       buffer_address <= {buffer_row, buffer_col};
	       buffer_state <= buffer_loop2;
	    end

	  buffer_loop2:
	    buffer_state <= buffer_loop2b;
	  buffer_loop2b:
	    buffer_state <= buffer_loop3;
	  buffer_loop3:
	    begin
	       vga_row_buffer_next[buffer_col] <= buffer_mem_out;
	       buffer_col <= buffer_col + 3'd1;
	       if(iVGA_request_y[8:0] == buffer_last_vga_row)
		 buffer_state <= buffer_loop1;
	       else
		 buffer_state <= row_change;
	    end
	endcase // case (buffer_state)
     end // always @ (posedge clk)
   	
   // Main Data Unit (Dualported M4K)
   datablock datamem (.address_a(mem_address),
		     .address_b(buffer_address),
		     .clock(clk),
		     .data_a(mem_data_in),
		     .data_b(1'b0),
		     .wren_a(mem_we),
		     .wren_b(1'b0),
		     .q_a(mem_data_out),
		     .q_b(buffer_mem_out));

   //////////////////////////////
   // Begin Main State Machine
   //////////////////////////////
   
   parameter wait_for_sync = 5'd0,
     start0 = 5'd1,
     start1 = 5'd2,
     start2 = 5'd16,
     save_middle_left = 5'd3,
     save_lower_left = 5'd4,
     save_middle_right = 5'd5,
     save_lower_right = 5'd6,
     load_neighbors = 5'd7,
     load_phase0 = 5'd8,
     load_phase1 = 5'd9,
     load_phase1b = 5'd15,
     load_phase2 = 5'd10,
     load_phase3 = 5'd11,
     calculate = 5'd12,
     save_phase = 5'd13,
     shift_phase = 5'd14;
   
   /*
   always @(posedge clk)
     begin
	mem_we <= 1'b1;
	mem_data_in <= 1'b1; //(iVGA_request_x[0]) ? 1'b1 : 1'b0;
	mem_address <= {iVGA_request_y[8:0], iVGA_request_x};
     end
    */
   
   always @(posedge clk)
     begin
	if(reset)
	  begin
	     col <= 3'd0;
	     state <= wait_for_sync;
	     mem_we <= 1'b1;  // Clear data upon reset
	     mem_data_in <= 1'b0;
	     mem_address <= {iVGA_request_y, iVGA_request_x};
	  end
	else if(ioverride_en)
	  begin
	     mem_we <= 1'b1;
	     mem_data_in <= ioverride_data;
	     mem_address <= {ioverride_y[8:0], ioverride_x};
	  end
	else
	  begin
	     case(state)
	       wait_for_sync:
		 begin
		    mem_we <= 1'b0;
		    //mem_we <= 1'b1;
		    //mem_data_in <= 1'b0; // Clear row below bottom (it will be read when updating last row)
		    //mem_address <= {HEIGHT, iVGA_request_x};
		    if(isync)
		      state <= start0;
		    row <= irow;
		    col <= 3'b0;
		 end

	       start0:
		 begin
		    if(row == 9'd0)
		      prev_row <= 8'd0;
		    else
		      prev_row <= original_row;

		    // Pipeline middle-left edge
		    mem_we <= 1'b0;
		    mem_address <= {row[8:0], 3'd0}; //WIDTH_WIRES

		    state <= start1;
		 end // case: start0

	       start1:
		 begin
		    // Load values from prev_row
		    oupper_left <= prev_row[0];
		    oupper_right <= prev_row[WIDTH-1];

		    // Preload upper value
		    upper_center <= prev_row[0];
		    
		    //Pipeline lower-left edge
		    mem_we <= 1'b0;
		    mem_address <= {row[8:0] + 9'd1, 3'd0}; //WIDTH_WIRES

		    state <= start2;
		 end // case: start1

	       start2:
		 begin
		    //Pipeline middle-right edge
		    mem_we <= 1'b0;
		    mem_address <= {row[8:0], WIDTH_VALUE}; //WIDTH_WIRES

		    state <= save_middle_left;
		 end

	       save_middle_left:
		 begin
		    omiddle_left <= mem_data_out;
		    middle_center <= mem_data_out; //col0 center value
		    
		    //Pipeline lower-right edge
		    mem_we <= 1'b0;
		    mem_address <= {row[8:0] + 9'd1, WIDTH_VALUE}; //WIDTH_WIRES

		    state <= save_lower_left;
		 end
	       
	       save_lower_left:
		 begin
		    olower_left <= mem_data_out;
		    lower_center <= mem_data_out;

		    state <= save_middle_right;
		 end

	       save_middle_right:
		 begin
		    omiddle_right <= mem_data_out;
		    state <= save_lower_right;
		 end
	       
	       save_lower_right:
		 begin
		    olower_right <= mem_data_out;
		    state <= load_neighbors;
		 end

	       load_neighbors:
		 begin
		    upper_left <= iupper_left;
		    middle_left <= imiddle_left;
		    lower_left <= ilower_left;
		    state <= load_phase0;
		 end

	       // END INITALIZATION; START LOOP
	       
	       load_phase0:
		 begin
		    mem_we <= 1'b0;
		    if(col == WIDTH_VALUE) begin
		       upper_right <= iupper_right;
		       middle_right <= imiddle_right;
		       lower_right <= ilower_right;
		       state <= calculate;
		    end
		    else begin
		       upper_right <= prev_row[col + 3'd1];
		       // Pipleine middle-right:
		       mem_address <= {row, col + 3'd1};
		       state <= load_phase1;
		    end // else: !if(col == WIDTH-1)
		 end // case: load_phase0

	       load_phase1:
		 begin
		    // Pipeline lower-right
		    mem_address <= {row + 9'd1, col + 3'd1};
		    state <= load_phase1b;
		 end

	       load_phase1b:
		 state <= load_phase2;

	       load_phase2:
		 begin
		    middle_right <= mem_data_out;
		    state <= load_phase3;
		 end
	       
	       load_phase3:
		 begin
		    lower_right <= mem_data_out;
		    state <= calculate;
		 end

	       calculate:
		 begin
		    original_row[col] <= middle_center;
		    if(neighbor_count == 4'd3)
		      mem_data_in <= 1'd1;
		    else if(neighbor_count == 4'd2 && middle_center)
		      mem_data_in <= 1'd1;
		    else
		      mem_data_in <= 1'd0;
		    mem_address <= {row, col};
		    state <= save_phase;
		 end

	       save_phase:
		 begin
		    mem_we <= 1'b1;
		    if(col == WIDTH_VALUE)
		      state <= wait_for_sync;
		    else
		      state <= shift_phase;
		 end // case: save_phase

	       shift_phase:
		 begin
		    mem_we <= 1'b0;
		    
		    // Increment col
		    col <= col + 3'd1;

		    // Fill left column
		    upper_left <= upper_center;
		    middle_left <= middle_center;
		    lower_left <= lower_center;

		    // Fill Middle column
		    upper_center <= upper_right;
		    middle_center <= middle_right;
		    lower_center <= lower_right;
		    
		    state <= load_phase0;
		 end // case: shift_phase

	     endcase // case (state)
	  end // else: !if(ioverride_en)
     end // always @ (posedge clk)
		 
   //state machine states:
   // reset (not actually a state) - current col becomes 0, state goes to wait_for_sync
   // wait_for_sync - waits for sync pulse to start row calculation
   // start - save row to reg, begin loading upper-left; if row==0, reset prev_row; col = 0
   // start2 - begin loading middle-left
   // save_upper_left - pipeline lower left
   // save_middle_left - pipeline upper right
   // save_lower_left - pipeline middle right
   // save_upper_right - pipeline lower right
   // save_middle_right - pipeline value at column 0
   // save_lower_right -pipeline value below column 0
   // 
   // init_phase1 - save column 0, prep for chain progression & calculation
   //             - Also save left neighbor values
   // init_phase2 - save value below column 0, prep for chain progression & calculation
   // load_phase1 - pipeline next
   // load_phase2 - pipeline next_lower
   // load_phase3 - save next
   // load_phase4 - save next_lower
   // calculate - Do calculation, save result
   // shift_phase1 - check if at end, if at end, return to wait_for_sync, else col++
   //              - Don't forget to disable we
   // shift_phase2 - middle_left_value <= current_value; 
   //                current_value <= middle_right_value;
   //                lower_left_value <= lower_center_value;
   //                lower_center_value <= lower_right_value;
   //             >> load_phase1
   
		 
endmodule // CellBlock

// data memory: 
// uses m4k blocks to hold a row of information
// adds an extra pixel on either end for buffer purposes.
module pixel_row (out, address, data, we, clk);
  output  [0:0] out;
  input [0:0] data;
  input [9:0] address;
  input we, clk;
  reg [9:0] read_add;
  reg [0:0] mem [641:0]; //extra pixels
  wire shifted_address = address + 10'd1;
  always @ (posedge clk) begin
     if (we) 
       mem[shifted_address] <= data;
     read_add <= shifted_address;
  end
  assign out = mem[read_add];
endmodule 
