module DE2_TOP (
		// Clock Input
		input 	      CLOCK_27, // 27 MHz
		input 	      CLOCK_50, // 50 MHz
		input 	      EXT_CLOCK, // External Clock
		// Push Button
		input [3:0]   KEY, // Pushbutton[3:0]
		// DPDT Switch
		input [17:0]  SW, // Toggle Switch[17:0]
		// 7-SEG Display
		output [6:0]  HEX0, // Seven Segment Digit 0
		output [6:0]  HEX1, // Seven Segment Digit 1
		output [6:0]  HEX2, // Seven Segment Digit 2
		output [6:0]  HEX3, // Seven Segment Digit 3
		output [6:0]  HEX4, // Seven Segment Digit 4
		output [6:0]  HEX5, // Seven Segment Digit 5
		output [6:0]  HEX6, // Seven Segment Digit 6
		output [6:0]  HEX7, // Seven Segment Digit 7
		// LED
		output [8:0]  LEDG, // LED Green[8:0]
		output [17:0] LEDR, // LED Red[17:0]
		// UART
		output 	      UART_TXD, // UART Transmitter
		input 	      UART_RXD, // UART Receiver
		// IRDA
		output 	      IRDA_TXD, // IRDA Transmitter
		input 	      IRDA_RXD, // IRDA Receiver
		// SDRAM Interface
		inout [15:0]  DRAM_DQ, // SDRAM Data bus 16 Bits
		output [11:0] DRAM_ADDR, // SDRAM Address bus 12 Bits
		output 	      DRAM_LDQM, // SDRAM Low-byte Data Mask 
		output 	      DRAM_UDQM, // SDRAM High-byte Data Mask
		output 	      DRAM_WE_N, // SDRAM Write Enable
		output 	      DRAM_CAS_N, // SDRAM Column Address Strobe
		output 	      DRAM_RAS_N, // SDRAM Row Address Strobe
		output 	      DRAM_CS_N, // SDRAM Chip Select
		output 	      DRAM_BA_0, // SDRAM Bank Address 0
		output 	      DRAM_BA_1, // SDRAM Bank Address 0
		output 	      DRAM_CLK, // SDRAM Clock
		output 	      DRAM_CKE, // SDRAM Clock Enable
		// Flash Interface
		inout [7:0]   FL_DQ, // FLASH Data bus 8 Bits
		output [21:0] FL_ADDR, // FLASH Address bus 22 Bits
		output 	      FL_WE_N, // FLASH Write Enable
		output 	      FL_RST_N, // FLASH Reset
		output 	      FL_OE_N, // FLASH Output Enable
		output 	      FL_CE_N, // FLASH Chip Enable
		// SRAM Interface
		inout [15:0]  SRAM_DQ, // SRAM Data bus 16 Bits
		output [17:0] SRAM_ADDR, // SRAM Address bus 18 Bits
		output 	      SRAM_UB_N, // SRAM High-byte Data Mask 
		output 	      SRAM_LB_N, // SRAM Low-byte Data Mask 
		output 	      SRAM_WE_N, // SRAM Write Enable
		output 	      SRAM_CE_N, // SRAM Chip Enable
		output 	      SRAM_OE_N, // SRAM Output Enable
		// ISP1362 Interface
		inout [15:0]  OTG_DATA, // ISP1362 Data bus 16 Bits
		output [1:0]  OTG_ADDR, // ISP1362 Address 2 Bits
		output 	      OTG_CS_N, // ISP1362 Chip Select
		output 	      OTG_RD_N, // ISP1362 Write
		output 	      OTG_WR_N, // ISP1362 Read
		output 	      OTG_RST_N, // ISP1362 Reset
		output 	      OTG_FSPEED, // USB Full Speed, 0 = Enable, Z = Disable
		output 	      OTG_LSPEED, // USB Low Speed,  0 = Enable, Z = Disable
		input 	      OTG_INT0, // ISP1362 Interrupt 0
		input 	      OTG_INT1, // ISP1362 Interrupt 1
		input 	      OTG_DREQ0, // ISP1362 DMA Request 0
		input 	      OTG_DREQ1, // ISP1362 DMA Request 1
		output 	      OTG_DACK0_N, // ISP1362 DMA Acknowledge 0
		output 	      OTG_DACK1_N, // ISP1362 DMA Acknowledge 1
		// LCD Module 16X2
		inout [7:0]   LCD_DATA, // LCD Data bus 8 bits
		output 	      LCD_ON, // LCD Power ON/OFF
		output 	      LCD_BLON, // LCD Back Light ON/OFF
		output 	      LCD_RW, // LCD Read/Write Select, 0 = Write, 1 = Read
		output 	      LCD_EN, // LCD Enable
		output 	      LCD_RS, // LCD Command/Data Select, 0 = Command, 1 = Data
		// SD Card Interface
		inout 	      SD_DAT, // SD Card Data
		inout 	      SD_DAT3, // SD Card Data 3
		inout 	      SD_CMD, // SD Card Command Signal
		output 	      SD_CLK, // SD Card Clock
		// I2C
		inout 	      I2C_SDAT, // I2C Data
		output 	      I2C_SCLK, // I2C Clock
		// PS2
		input 	      PS2_DAT, // PS2 Data
		input 	      PS2_CLK, // PS2 Clock
		// USB JTAG link
		input 	      TDI, // CPLD -> FPGA (data in)
		input 	      TCK, // CPLD -> FPGA (clk)
		input 	      TCS, // CPLD -> FPGA (CS)
		output 	      TDO, // FPGA -> CPLD (data out)
		// VGA
		output 	      VGA_CLK, // VGA Clock
		output 	      VGA_HS, // VGA H_SYNC
		output 	      VGA_VS, // VGA V_SYNC
		output 	      VGA_BLANK, // VGA BLANK
		output 	      VGA_SYNC, // VGA SYNC
		output [9:0]  VGA_R, // VGA Red[9:0]
		output [9:0]  VGA_G, // VGA Green[9:0]
		output [9:0]  VGA_B, // VGA Blue[9:0]
		// Ethernet Interface
		inout [15:0]  ENET_DATA, // DM9000A DATA bus 16Bits
		output 	      ENET_CMD, // DM9000A Command/Data Select, 0 = Command, 1 = Data
		output 	      ENET_CS_N, // DM9000A Chip Select
		output 	      ENET_WR_N, // DM9000A Write
		output 	      ENET_RD_N, // DM9000A Read
		output 	      ENET_RST_N, // DM9000A Reset
		input 	      ENET_INT, // DM9000A Interrupt
		output 	      ENET_CLK, // DM9000A Clock 25 MHz
		// Audio CODEC
		inout 	      AUD_ADCLRCK, // Audio CODEC ADC LR Clock
		input 	      AUD_ADCDAT, // Audio CODEC ADC Data
		inout 	      AUD_DACLRCK, // Audio CODEC DAC LR Clock
		output 	      AUD_DACDAT, // Audio CODEC DAC Data
		inout 	      AUD_BCLK, // Audio CODEC Bit-Stream Clock
		output 	      AUD_XCK, // Audio CODEC Chip Clock
		// TV Decoder
		input [7:0]   TD_DATA, // TV Decoder Data bus 8 bits
		input 	      TD_HS, // TV Decoder H_SYNC
		input 	      TD_VS, // TV Decoder V_SYNC
		output 	      TD_RESET, // TV Decoder Reset
		// GPIO
		inout [35:0]  GPIO_0, // GPIO Connection 0
		inout [35:0]  GPIO_1       // GPIO Connection 1
		);

   //Turn off all displays.
   assign HEX0 = 7'h7F;
   assign HEX1 = 7'h7F;
   assign HEX2 = 7'h7F;
   assign HEX3 = 7'h7F;
   assign HEX4 = 7'h7F;
   assign HEX5 = 7'h7F;
   assign HEX6 = 7'h7F;
   assign HEX7 = 7'h7F;
   //   assign LEDR = 18'h0;
   // assign LEDG = 9'h0;
   
   //Set all GPIO to tri-state.
   assign GPIO_0 = 36'hzzzzzzzzz;
   assign GPIO_1 = 36'hzzzzzzzzz;

   //Disable audio codec.
   //assign AUD_DACDAT = 1'b0;
   //assign AUD_XCK    = 1'b0;

   //Disable DRAM.
   assign DRAM_ADDR  = 12'h0;
   assign DRAM_BA_0  = 1'b0;
   assign DRAM_BA_1  = 1'b0;
   assign DRAM_CAS_N = 1'b1;
   assign DRAM_CKE   = 1'b0;
   assign DRAM_CLK   = 1'b0;
   assign DRAM_CS_N  = 1'b1;
   assign DRAM_DQ    = 16'hzzzz;
   assign DRAM_LDQM  = 1'b0;
   assign DRAM_RAS_N = 1'b1;
   assign DRAM_UDQM  = 1'b0;
   assign DRAM_WE_N  = 1'b1;

   //Disable Ethernet.
   assign ENET_CLK   = 1'b0;
   assign ENET_CS_N  = 1'b1;
   assign ENET_CMD   = 1'b0;
   assign ENET_DATA  = 16'hzzzz;
   assign ENET_RD_N  = 1'b1;
   assign ENET_RST_N = 1'b1;
   assign ENET_WR_N  = 1'b1;

   //Disable flash.
   assign FL_ADDR  = 22'h0;
   assign FL_CE_N  = 1'b1;
   assign FL_DQ    = 8'hzz;
   assign FL_OE_N  = 1'b1;
   assign FL_RST_N = 1'b1;
   assign FL_WE_N  = 1'b1;

   //Disable LCD.
   assign LCD_BLON = 1'b0;
   assign LCD_DATA = 8'hzz;
   assign LCD_EN   = 1'b0;
   assign LCD_ON   = 1'b0;
   assign LCD_RS   = 1'b0;
   assign LCD_RW   = 1'b0;

   //Disable OTG.
   assign OTG_ADDR    = 2'h0;
   assign OTG_CS_N    = 1'b1;
   assign OTG_DACK0_N = 1'b1;
   assign OTG_DACK1_N = 1'b1;
   assign OTG_FSPEED  = 1'b1;
   assign OTG_DATA    = 16'hzzzz;
   assign OTG_LSPEED  = 1'b1;
   assign OTG_RD_N    = 1'b1;
   assign OTG_RST_N   = 1'b1;
   assign OTG_WR_N    = 1'b1;

   //Disable SDRAM.
   assign SD_DAT = 1'bz;
   assign SD_CLK = 1'b0;

   //Disable SRAM.
   assign SRAM_ADDR = 18'h0;
   assign SRAM_CE_N = 1'b1;
   assign SRAM_DQ   = 16'hzzzz;
   assign SRAM_LB_N = 1'b1;
   assign SRAM_OE_N = 1'b1;
   assign SRAM_UB_N = 1'b1;
   assign SRAM_WE_N = 1'b1;

   //Disable VGA.
   /*
    assign VGA_CLK   = 1'b0;
    assign VGA_BLANK = 1'b0;
    assign VGA_SYNC  = 1'b0;
    assign VGA_HS    = 1'b0;
    assign VGA_VS    = 1'b0;
    assign VGA_R     = 10'h0;
    assign VGA_G     = 10'h0;
    assign VGA_B     = 10'h0;
    */

   //Disable all other peripherals.
   //assign I2C_SCLK = 1'b0;
   assign IRDA_TXD = 1'b0;
   //assign TD_RESET = 1'b0;
   assign TDO = 1'b0;
   assign UART_TXD = 1'b0;
   

   wire 		      VGA_CTRL_CLK;
   wire 		      AUD_CTRL_CLK;
   wire 		      DLY_RST;
   
   assign  TD_RESET  =  1'b1;  //  Allow 27 MHz
   assign  AUD_ADCLRCK  =  AUD_DACLRCK;
   assign  AUD_XCK    =  AUD_CTRL_CLK;

   Reset_Delay      r0  (  .iCLK(CLOCK_50),.oRESET(DLY_RST)  );
   
   VGA_Audio_PLL     p1  (  .areset(~DLY_RST),.inclk0(CLOCK_27),.c0(VGA_CTRL_CLK),.c1(AUD_CTRL_CLK),.c2(VGA_CLK)  );
   
   
   VGA_Controller    u1  (  //  Host Side
			    .iCursor_RGB_EN(4'b1111),
			    .oAddress(mVGA_ADDR),
			    .oCoord_X(Coord_X),
			    .oCoord_Y(Coord_Y),
			    .iCursor_X(SW[17:11]),
			    .iCursor_Y(SW[10:4]),
			    .iCursor_R(10'h3FF),
			    .iCursor_G(10'h3FF),
			    .iCursor_B(10'h0),
			    .iRed(mVGA_R),
			    .iGreen(mVGA_G),
			    .iBlue(mVGA_B),
			    //  VGA Side
			    .oVGA_R(VGA_R),
			    .oVGA_G(VGA_G),
			    .oVGA_B(VGA_B),
			    .oVGA_H_SYNC(VGA_HS),
			    .oVGA_V_SYNC(VGA_VS),
			    .oVGA_SYNC(VGA_SYNC),
			    .oVGA_BLANK(VGA_BLANK),
			    //  Control Signal
			    .iCLK(VGA_CTRL_CLK),
			    .iRST_N(DLY_RST)  );

   wire [9:0] 		      mVGA_R;        //memory output to VGA
   wire [9:0] 		      mVGA_G;
   wire [9:0] 		      mVGA_B;
   wire [18:0] 		      mVGA_ADDR;      //video memory address
   wire [9:0] 		      Coord_X, Coord_Y;  //display coods

   ////////////////////////////////////
   //DLA state machine variables
   wire 		      reset;
   reg [3:0] 		      state;  //state machine
   reg [30:0] 		      x_rand;  //shift registers for random number gen  
   reg [28:0] 		      y_rand;
   wire 		      seed_low_bit, x_low_bit, y_low_bit; //rand low bits for SR
   //reg [9:0] x_walker; //particle coords of random walker
   //reg [8:0] y_walker;
   reg [3:0] 		      sum; //neighbor sum

   assign LEDG = sum;

   ////////////////////////////////////
   // My Lab1 Variables:
   reg [9:0] 		      col;  // Column and Row for current calculation location
   reg [8:0] 		      row;
   reg [7:0] 		      rule;  // rule for calculations

   // Temporary stores for values from previous row
   reg 			      left;
   reg 			      center;
   // don't need a right since it'll be the last one read from memory.

   assign LEDR[7:0] = rule;  // Pretty LEDs for Rule Selection
   assign LEDR[17:14] = state;
   assign LEDR[13:10] = neighbor_count;
   assign LEDR[9] = result;
   assign LEDR[8] = middle_center;
      
   wire 		      user_button;
   assign user_button = ~KEY[3];
   wire [18:0] 		      display_address;
   wire [18:0] 		      prev_left_address;
   wire [18:0] 		      prev_center_address;
   wire [18:0] 		      prev_right_address;
   assign display_address = 19'd480*Coord_X[9:0] + Coord_Y[8:0]; // vga current address

   reg 			      blue_en1;
   reg 			      green_en1;
   reg 			      red_en1;
   reg 			      blue_en2;
   reg 			      green_en2;
   reg 			      red_en2;
   
   //buffer for edges of VGA out
   parameter COL_START = 10'd0;
   parameter COL_END = 10'd639;
   parameter ROW_START = 9'd0;
   parameter ROW_END = 10'd479;
   
   ////////////////////////////////////
   /*From megaWizard:
    module vga_buffer (
    address_a, // use a for state machine
    address_b, // use b for VGA refresh
    clock_a,
    clock_b,
    data_a,
    data_b,
    wren_a,
    wren_b,
    q_a,
    q_b);*/
   // Show m4k on the VGA
   // -- use m4k a for state machine
   // -- use m4k b for VGA refresh
   wire 		      mem_bit ; //current data from m4k to VGA
   wire 			      disp_bit ; // registered data from m4k to VGA
   wire 		      state_bit ; // current data from m4k to state machine
   reg 			      we ; // write enable for a
   reg [18:0] 		      addr_reg ; // for a
   reg 			      data_reg ; // for a
   vga_buffer display(
		      .address_a (addr_reg) , 
		      .address_b (display_address), // vga current address
		      .clock_a (VGA_CTRL_CLK),
		      .clock_b (VGA_CTRL_CLK),
		      .data_a (data_reg),
		      .data_b (1'b0), // never write on port b
		      .wren_a (we),
		      .wren_b (1'b0), // never write on port b
		      .q_a (state_bit),
		      .q_b (mem_bit) ); // data used to update VGA

   // make the color
   assign  mVGA_R = {10{disp_bit}}*red_en2 - {10{disp_bit}}*Coord_Y*10'd2*red_en1;
   assign  mVGA_G = {10{disp_bit}}*green_en2 - {10{disp_bit}}*Coord_Y*10'd2*green_en1;
   assign  mVGA_B = {10{disp_bit}}*blue_en2 - {10{disp_bit}}*Coord_Y*10'd2*blue_en1;


   // DLA state machine
   assign reset = ~KEY[0];

   //right-most bit for rand number shift regs
   //your basic XOR random # gen
   assign x_low_bit = x_rand[27] ^ x_rand[30];
   assign y_low_bit = y_rand[26] ^ y_rand[28];

   reg [18:0] 		      prev_pixel;

   //Grid Locations
   reg [0:0] 		      upper_left;
   reg [0:0] 		      upper_center;
   reg [0:0] 		      upper_right;
   reg [0:0] 		      middle_left;
   reg [0:0] 		      middle_center;
   reg [0:0] 		      middle_right;
   reg [0:0] 		      lower_left;
   reg [0:0] 		      lower_center;
   reg [0:0] 		      lower_right;  

   reg [0:0] 		      middle_right_set;
   reg [0:0] 		      lower_left_set;
   reg [0:0] 		      lower_center_set;
   reg [0:0] 		      lower_right_set;

   reg [0:0] 		      result;
   
   
   wire [3:0] 		      neighbor_count;
   assign neighbor_count = upper_left + 
			   upper_center + 
			   upper_right +  
			   middle_left +  
			   middle_right + 
			   lower_left +   
			   lower_center + 
			   lower_right;  

   //state names
   parameter init = 4'd0, 
     update_load1 = 4'd1,
     update_load2 = 4'd2,
     update_load3 = 4'd3,
     update_load4 = 4'd4,
     update_load5 = 4'd5,
     update_load6 = 4'd6,
     update_calculate = 4'd7,
     wait_for_vga_start = 4'd8,
     wait_for_next_pixel = 4'd9,
     save_result = 4'd10,
     save_result2 = 4'd11;
   /*
   always @ (negedge VGA_CTRL_CLK)
     begin
	// register the m4k output for better timing on VGA
	// negedge seems to work better than posedge
	disp_bit <= mem_bit;
     end
    */
   wire prev_row_out;
   reg [0:0] prev_row_data;
   reg [9:0] prev_row_address;
   reg [0:0] prev_row_we;
   
   pixel_row prev_row(.out(prev_row_out),
		      .address(prev_row_address),
		      .data(prev_row_data),
		      .we(prev_row_we),
		      .clk(CLOCK_50));

   GridController controller(.iVGA_request_x(Coord_X),
			     .iVGA_request_y(Coord_Y),
			     .oVGA_value(disp_bit),
			     .clk(VGA_CTRL_CLK));
   
   
   
   always @ (posedge VGA_CTRL_CLK) //VGA_CTRL_CLK
     begin
	// register the m4k output for better timing on VGA
	//disp_bit <= mem_bit;
	x_rand <= {x_rand[29:0] , x_low_bit} ;
	
	if (reset) //synch reset assumes KEY0 is held down 1/60 second
	  begin
	     //clear the screen
	     addr_reg <= display_address;
	     we <= 1'b1;          //write some memory
	     data_reg <= x_rand[0];        //write random bits
	     //clear prev_row
	     prev_row_we <= 1'b1;
	     prev_row_data <= 1'b0;
	     prev_row_address <= display_address[9:0];
	     //init random number generators to alternating bits
	     if (x_rand == 31'h0)
               x_rand <= 31'h55555555;
	     y_rand <= 29'h55555555;
	     state <= init;  //first state in regular state machine 
	     // Pick Random Colors:
	     red_en1 <= x_rand[0];
	     red_en2 <= x_rand[1];
	     green_en1 <= x_rand[2];
	     green_en2 <= x_rand[3];
	     blue_en1 <= x_rand[4];
	     blue_en2 <= x_rand[5];
	  end // if (reset)
	/*
	else
	  begin
	     case(state)
               init:
		 begin
		    prev_row_we <= 1'b0;
		    state <= wait_for_vga_start;
		 end // case: init

	       wait_for_vga_start:
		 begin
		    we <= 1'b0;
		    if(Coord_X[9:0] == 10'd0 && Coord_Y[8:0] == 9'd0)
		      state <= update_load1;
		 end

	       update_load1:
		 begin
		    // House keeping:
		    prev_pixel <= display_address;
		    lower_left_set <= 1'b0;
		    lower_center_set <= 1'b0;
		    lower_right_set <= 1'b0;

		    // Retrive saved previous
		    if(Coord_X[9:0] == 10'd0)
		      middle_left <= 1'b0;
		    else
		      middle_left <= result;
		    
		    // Pipeline middle_right
		    we <= 1'd0;
		    if(Coord_X[9:0] == 10'd639)
		      begin
			 middle_right_set <= 1'b1;
			 middle_right <= 1'b0;
		      end
		    else
		      begin
			 middle_right_set <= 1'b0;
			 addr_reg <= display_address + 'd1;
		      end

		    //Pipeline upper_left
		    prev_row_address <= Coord_X[9:0] - 10'd1;
		    prev_row_we <= 1'b0;

		    state <= update_load2;
		 end // case: update_load1
	       	       
		update_load2:
		  begin
		     //Pipeline lower_left
		     if(Coord_X[9:0] == 10'd0 || Coord_Y[8:0] == 9'd479) begin
			lower_left_set <= 1'b1;
			lower_left <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd639;
		     
		     //Pipeline upper_center
		     prev_row_address <= Coord_X[9:0];

		     //Save upper_left from Pipeline
		     upper_left <= prev_row_out;

		     state <= update_load3;
		  end // case: update_load2
	       
		update_load3:
		  begin
		     //Pipeline lower_center
		     if(Coord_Y[8:0] == 9'd479) begin
			lower_center_set <= 1'b1;
			lower_center <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd640;

		     //Save middle_right from pipeline
		     if(!middle_right_set)
		       middle_right <= state_bit;
		     
		     //Pipeline upper_right
		     prev_row_address <= Coord_X[9:0] + 10'd1;

		     //Save upper_center from Pipeline
		     upper_center <= prev_row_out;

		     state <= update_load4;
		  end // case: update_load3
	       
		update_load4:
		  begin
		     //load center:
		     middle_center <= disp_bit;
		     
		     //Pipeline lower_right
		     if(Coord_Y[8:0] == 9'd479 || Coord_X[9:0] == 10'd639) begin
			lower_right_set <= 1'b1;
			lower_right <= 1'b0;
		     end
		     else
		       addr_reg <= display_address + 'd641;

		     //Save lower_left from pipeline
		     if(!lower_left_set)
		       lower_left <= state_bit;
		     
		     //Save upper_right from Pipeline
		     upper_right <= prev_row_out;

		     state <= update_load5;
		  end // case: update_load4
	       	       
		update_load5:
		  begin
		     //Save lower_center from pipeline
		     if(!lower_center_set)
		       lower_center <= state_bit;

		     state <= update_load6;
		  end // case: update_load5
	       	       
		update_load6:
		  begin
		     //Save lower_center from pipeline
		     if(!lower_right_set)
		       lower_right <= state_bit;

		     state <= update_calculate;
		  end // case: update_load6
	       
	       update_calculate:
		 begin
		    if(neighbor_count == 'd3)
		      result <= 1'd1;
		    else if(neighbor_count == 'd2 && middle_center)
		      result <= 1'd1;
		    else
		      result <= 1'd0;
		    state <= wait_for_next_pixel;

		    // Also, update prev_row:
		    prev_row_we <= 1'b1;
		    prev_row_data <= middle_left;
		    prev_row_address <= Coord_X[9:0] - 'd1;
		 end

	       wait_for_next_pixel:
		 begin
		    if(prev_pixel != display_address)
		      state <= save_result;
		 end

	       save_result:
		 begin
		    addr_reg <= prev_pixel;
		    we <= 1'b1;
		    data_reg <= result;
		    state <= save_result2;
		 end
	       
	       save_result2:
		 begin
		    if(KEY[3]) begin
		       if(Coord_X == 10'd0 && Coord_Y == 9'd0)
			 state <= wait_for_vga_start;
		       else
			 state <= update_load1;
		    end
		 end // case: save_result2
	     endcase // case (state)
	  end // else: !if(reset) */
     end // always @ (posedge VGA_CTRL_CLK)
endmodule //top module

module GridController(iVGA_request_x,  
		      iVGA_request_y,  
		      irun,	       
		      ioverride_x,     
		      ioverride_y,     
		      ioverride_data,  
		      ioverride_clk,   
		      clk,	       
		      reset,	       
		      oVGA_value,      
		      orunning
		      );
   
   input [9:0] iVGA_request_x;  
   input [9:0] iVGA_request_y;  
   input       irun;	       
   input [9:0] ioverride_x;     
   input [9:0] ioverride_y;     
   input       ioverride_data;  
   input       ioverride_clk;   
   input       clk;	       
   input       reset;	       
   output      oVGA_value;      
   output      orunning;   

   parameter CELLBLOCKS = 80;

   // Interconnects between cell blocks
   wire interconnects[CELLBLOCKS+1][6];
   wire VGA_outs[CELLBLOCKS];
   wire row;
   wire sync;
   
   
   genvar i;
   generate
      for (i=0; i < CELLBLOCKS; i=i+1) begin : GENLOOP
	    CellBlock block (.iVGA_request_x(iVGA_request_x[2:0]),
			     .iVGA_request_y(iVGA_request_y),
			     .irow(row),		 
			     .isync(sync),		 
			     .clk(clk),		 
			     .ioverride_en(),	 
			     .ioverride_x(),	 
			     .ioverride_y(),	 
			     .ioverride_data(), 
			     .iupper_right((i == CELLBLOCKS-1) ? 1'b0 : interconnects[i+1][3]),   
			     .imiddle_right((i == CELLBLOCKS-1) ? 1'b0 : interconnects[i+1][4]),	 
			     .ilower_right((i == CELLBLOCKS-1) ? 1'b0 : interconnects[i+1][5]),	 
			     .iupper_left((i == 0) ? 1'b0 : interconnects[i][0]),	 
			     .imiddle_left((i == 0) ? 1'b0 : interconnects[i][1]),	 
			     .ilower_left((i == 0) ? 1'b0 : interconnects[i][2]),            	 
			     .oVGA_value(VGA_outs[i]),	 
			     .oupper_right(interconnects[i+1][0]),  
			     .omiddle_right(interconnects[i+1][1]), 
			     .olower_right(interconnects[i+1][2]),  
			     .oupper_left(interconnects[i][3]),	       
			     .omiddle_left(interconnects[i][4]),	       
			     .olower_left(interconnects[i][5])
		    );
      end // block: GENLOOP
   endgenerate

   assign oVGA_value = VGA_outs[iVGA_request_x[9:3]];
   
   
   /*
   always @(posedge clk)
     begin
	

	oVGA_value <= iVGA_request_x[0] ^ iVGA_request_y[0];
     end
   */

endmodule // GridController

module CellBlock(iVGA_request_x,
		 iVGA_request_y,
		 irow,		
		 isync,		
		 clk,		
		 ioverride_en,	
		 ioverride_x,	
		 ioverride_y,	
		 ioverride_data,
		 iupper_right,  
		 imiddle_right,	
		 ilower_right,	
		 iupper_left,	
		 imiddle_left,	
		 ilower_left,	
		 oVGA_value,	
		 oupper_right,  
		 omiddle_right,	
		 olower_right,	
		 oupper_left,	
		 omiddle_left,	
		 olower_left     );

   input [2:0] iVGA_request_x;
   input [9:0] iVGA_request_y;
   input irow;		
   input isync;		
   input clk;		
   input ioverride_en;	
   input [2:0] ioverride_x;	
   input [9:0] ioverride_y;	
   input ioverride_data;
   input iupper_right;  
   input imiddle_right;	
   input ilower_right;	
   input iupper_left;	
   input imiddle_left;	
   input ilower_left;	
   output oVGA_value;	
   output oupper_right;  
   output omiddle_right;	
   output olower_right;	
   output oupper_left;	
   output omiddle_left;	
   output olower_left;
	
   reg 	  mem_we, mem_data, mem_address;
   wire   mem_out;
   
   datablock datamem (.address_a(mem_address),
		     .address_b({iVGA_request_y[8:0], iVGA_request_x}),
		     .clock(clk),
		     .data_a(mem_data),
		     .data_b(1'b0),
		     .wren_a(mem_we),
		     .wren_b(1'b0),
		     .q_a(mem_out),
		     .q_b(oVGA_value));

   always @(posedge clk)
     begin
	mem_we <= 1'b1;
	mem_data <= 1'b1; //(iVGA_request_x[0]) ? 1'b1 : 1'b0;
	mem_address <= {iVGA_request_y[8:0], iVGA_request_x};
     end
   
   
		 
endmodule // CellBlock

// data memory: 
// uses m4k blocks to hold a row of information
// adds an extra pixel on either end for buffer purposes.
module pixel_row (out, address, data, we, clk);
  output  [0:0] out;
  input [0:0] data;
  input [9:0] address;
  input we, clk;
  reg [9:0] read_add;
  reg [0:0] mem [641:0]; //extra pixels
  wire shifted_address = address + 10'd1;
  always @ (posedge clk) begin
     if (we) 
       mem[shifted_address] <= data;
     read_add <= shifted_address;
  end
  assign out = mem[read_add];
endmodule 
